;;;;;; FASL library for Scheme48                         -*- Scheme -*-
;;;;;; User environment commands

;;; Taylor Campbell wrote this code; he places it in the public domain.

(define-syntax define-command
  (syntax-rules ()
    ((DEFINE-COMMAND name syntax help args
       procedure)
     (BEGIN (DEFINE-USER-COMMAND-SYNTAX (QUOTE name) syntax
              help
              (QUOTE args))
            (ENVIRONMENT-DEFINE! (USER-COMMAND-ENVIRONMENT)
                                 (QUOTE name)
                                 procedure)))))

(define-command dump-package "<package> [<filename>]"
  "compile and dump a package to the disk"
  (name &opt filename)
  (lambda (name . maybe-filename)
    (compile-and-dump-package-to-file
     (get-package name)
     (translate (if (null? maybe-filename)
                    (string-append (symbol->string name) ".sco")
                    (car maybe-filename))))))

(define load-package-from-source
        (environment-ref (user-command-environment)
                         'load-package))

(define-command load-package "<package> [<compiled-filename>]"
  "load a package, possibly from a compiled dump"
  (name &opt filename)
  (lambda (name . maybe-filename)
    (let* ((package (get-package name))
           (filename (if (null? maybe-filename)
                         (string-append (file-name-directory
                                         (package-file-name package))
                                        (symbol->string name)
                                        ".sco")
                         (car maybe-filename)))
           (filename (and filename (translate filename))))
      (if (not (and filename (file-exists? filename)))
          (load-package-from-source name)
          (retrieve-and-load-package-from-file package filename)))))

(define (file-exists? filename)
  ;++ Horrid hack.
  (call-with-current-continuation
    (lambda (k)
      (with-handler (lambda (c punt) (k #f))
        (lambda ()
          (close-input-port (open-input-file filename))
          #t)))))

(define (get-package name)
  (let ((package (structure-package (get-structure name))))
    (if (package-unstable? package)
        package
        (error "can't dump or load stable package"
               name))))

(define-command dump-fasl "<filename> (<struct> ...) <aux-obj> ..."
  "dump a FASL"
  (filename form &rest form)
  (lambda (filename structs . aux-objs)
    (let ((structs
           (map (lambda (struct)
                  (receive (name exp)
                           (if (pair? struct)
                               (values (car struct)
                                       ;; Support both (name struct) &
                                       ;; (name . struct).
                                       (if (pair? (cdr struct))
                                           (cadr struct)
                                           (cdr struct)))
                               (values struct struct))
                    (cons name
                          (eval exp (config-package)))))
                structs))
          (aux-objs (map (lambda (aux)
                           (eval aux (config-package)))
                         aux-objs)))
      (dump-fasl (translate filename) structs aux-objs))))

(define-command load-fasl "<filename>"
  "load a FASL"
  (filename)
  (lambda (filename)
    (load-fasl (translate filename))))
