;;;;;; FASL library for Scheme48                         -*- Scheme -*-
;;;;;; Package compilation & dumping

;;; Taylor Campbell wrote this code; he places it in the public domain.

(define (compile-and-dump-package-to-file package filename)
  (compile-and-dump-package package
    (lambda (dump-encoder receiver)
      (call-with-file-dumpster filename dump-encoder receiver))))

(define (retrieve-and-load-package-from-file package filename)
  (retrieve-and-load-package package
    (lambda (dump-decoder receiver)
      (call-with-file-retriever filename dump-decoder receiver))))

(define package-dump-level 7)



;;; This is mostly duplicated from bcomp/comp-package.scm, because it
;;; is not _quite_ parameterized how we want it.  (We want to pass in
;;; our own compiler environment; it wants to just get PACKAGE's with
;;; PACKAGE->ENVIRONMENT.)

(define (compile-package package cenv)
  (let ((template (compile-forms ((get-optimizer
                                   (package-optimizer-names package))
                                  (expand-package package cenv)
                                  package)
                                 (package-name package)
                                 (package-uid package))))
    (link! template package #t)         ; #T -> warn about undefined
    template))

(define (expand-package package cenv)
  (receive (files+forms transforms needs-primitives?)
           (package-source package)
    (for-each (lambda (name)
                (define-usual-transform! package cenv name))
              transforms)
    (let ((cenvs+scanned
           (map (lambda (file.forms)
                  (let ((cenv (bind-source-file-name (car file.forms)
                                                     cenv)))
                    (cons cenv (scan-forms (cdr file.forms) cenv))))
                (if needs-primitives?
                    (cons (cons #f (define-primitives! cenv))
                          files+forms)
                    files+forms))))
      (reverse
       (fold (lambda (cenv.scanned expanded)
               (fold (let ((cenv (car cenv.scanned)))
                       (lambda (form expanded)
                         (cons (delay (expand-scanned-form form cenv))
                               expanded)))
                     (cdr cenv.scanned)
                     expanded))
             cenvs+scanned
             '())))))

(define (define-usual-transform! package cenv name)
  (environment-define! cenv name syntax-type
                       (make-transform (usual-transform name)
                                       package
                                       syntax-type
                                       `(USUAL-TRANSFORM ',name)
                                       name)))

(define (define-primitives! cenv)
  (table-walk (lambda (name op)
                (let ((type (operator-type op)))
                  (if (not (eq? type 'leaf))
                      (environment-define! cenv name type op))))
              operators-table)
  (environment-define! cenv 'ALL-OPERATORS vector-type)
  (let ((all-operators-node (expand 'ALL-OPERATORS cenv))
        (vector-set!-node
         (make-node op/literal (get-primop 'VECTOR-SET!)))
        (nodes '())
        (primop-count (let ((count 0))
                        (walk-primops (lambda (name type primop)
                                        (set! count (+ count 1))))
                        count))
        (index 0))
    (define (make-primitive-definition name)
      (make-node op/define
                 `(DEFINE ,(expand name cenv)
                    ,(make-node op/primitive-procedure
                                `(PRIMITIVE-PROCEDURE ,name)))))
    (define (make-primitive-registration name index)
      (make-node op/call
                 `(,vector-set!-node ,all-operators-node
                                     ,(make-node op/literal index)
                                     ,(expand name cenv))))
    (walk-primops (lambda (name type primop)
                    (environment-define! cenv name type primop)
                    (set! nodes
                          (cons (make-primitive-definition name)
                                (cons (make-primitive-registration
                                       name index)
                                      nodes)))
                    (set! index (+ index 1))))
    (cons (make-node op/define
            `(DEFINE ,all-operators-node
               ,(make-node op/call
                           `(,(make-node op/literal
                                         (get-primop 'MAKE-VECTOR))
                             ,(make-node op/literal primop-count)))))
          nodes)))

(define op/define              (get-operator 'DEFINE))
(define op/primitive-procedure (get-operator 'PRIMITIVE-PROCEDURE))
(define op/call                (get-operator 'CALL))
(define op/literal             (get-operator 'LITERAL))



(define (compile-and-dump-package package call-with-dumpster)
  (let* ((table (make-location-table))
         (cenv (make-cenv-for-dump table
                                   package
                                   (package->environment package)))
         (template (compile-package package cenv)))
    ;; This allows the dumpster to remain unopened until necessary.
    ;; Compiling the package can be expensive.
    (call-with-dumpster dump-encoder-for-packages
      (lambda (dumpster)
        (dump-compiled-package dumpster package table template cenv)))
    template))

(define (make-location-table) (make-table location-id))

; (define make-location-table
;         (table-constructor
;          (make-table-template location? eq? location-id
;                               #f #f #f
;                               'location-table)))

(define (make-cenv-for-dump table package cenv)
  (make-compiler-env (lambda (name)
                       (let ((probe (cenv name)))
                         (cache-binding name probe cenv table)
                         probe))
                     (lambda (name type . maybe-value)
                       (apply environment-define! cenv name type
                              maybe-value))
                     (make-tower-for-dump table package cenv)
                     package))

(define (cache-binding name probe cenv table)
  (if (binding? probe)
      (let ((type (binding-type probe))
            (loc (binding-place probe))
            (static (binding-static probe)))
        (cond ((and loc (not (table-ref table loc)))
               (table-set! table loc
                           (cons (name->qualified name cenv)
                                 (type->sexp type #t)))
               (if (transform? static)
                   (probe-aux-names name static cenv table)))))))

(define (probe-aux-names name static cenv table)
  (cond ((transform-aux-names static)
         => (lambda (aux-names)
              (let ((t-cenv (package->environment
                             (transform-env static))))
                (for-each (lambda (aux-name)
                            (cache-binding
                             (generate-name aux-name cenv name)
                             (t-cenv aux-name)
                             cenv
                             table))
                          aux-names))))
        (else
         (warn "transform without auxiliary names"
               static))))

(define (make-tower-for-dump table package cenv)
  (delay
    (let* ((tower (force (environment-macro-eval cenv)))
           (next-env (cdr tower)))
      (cons (car tower)
            (if (package? next-env)
                (let* ((next-cenv (package->environment next-env))
                       (replacement
                        (make-simple-package
                         (make-tower-opens next-env)
                         #t             ; unstable
                         (make-tower-for-dump table next-env
                                              next-cenv))))
                  (set-package->environment! replacement
                    (make-cenv-for-dump table replacement next-cenv))
                  replacement)
                ;; Not sure why it might not be a package, but just in
                ;; case...
                next-env)))))

(define (set-package->environment! package cenv)
  ;++ Hurk, spew, ralph.
  (record-set! package 10 cenv))

(define (make-tower-opens package)
  (let ((names '()))
    (table-walk (lambda (name binding)
                  (set! names (cons (list name (binding-type binding))
                                    names)))
                (package-definitions package))
    (cons (make-structure package
                          (make-simple-interface #f names))
          (package-opens package))))

(define (dump-compiled-package dumpster package table template cenv)
  (dump-integer dumpster package-dump-level)
  (let-fluids $location-table table
              $cenv-for-dump cenv
    (lambda ()
      (dump-bindings dumpster package cenv table)
      (dump-template dumpster template))))

(define $location-table (make-fluid #f))
(define $cenv-for-dump  (make-fluid #f))

(define (dump-bindings dumpster package cenv table)
  (table-walk (lambda (name binding)
                (dump dumpster name)
                (dump dumpster (type->sexp (binding-type binding) #t))
                (let ((static (binding-static binding)))
                  (cond ((transform? static)
                         (probe-aux-names name static cenv table)
                         (dump-transform dumpster static))
                        (else
                         (dump dumpster #f)))))
              (package-definitions package))
  (dump dumpster #f))

(define (dump-transform dumpster transform)
  (dump dumpster (transform-id transform))
  (dump dumpster (type->sexp (transform-type transform) #t))
  (let ((source (transform-source transform)))
    (if (and (pair? source)             ; Kludge.
             (eq? (car source) 'inline-transform))
        (destructure (( (inline-transform (quote0 source)
                                          (quote1 aux-names))
                        source))
          (dump dumpster inline-transform)
          (dump dumpster aux-names)
          (dump dumpster source))
        (begin (dump dumpster (transform-procedure transform))
               (dump dumpster (transform-aux-names transform))
               (dump dumpster source)))))

;;; Blah: TRANSFORMS doesn't export this, so we need to assume that
;;; transform records are laid out with the procedure in slot 1.

(define (transform-procedure transform)
  (record-ref transform 1))

(define (dump-encoder-for-packages obj)
  (cond ((primop?    obj) (values #t 'primop       primop-dump-encoder))
        ((generated? obj) (values #t 'generated generated-dump-encoder))
        ((closure?   obj) (values #f 'closure     closure-dump-encoder))
        (else             (usual-dump-encoder obj))))

(define primop-dump-encoder
  (lambda (dumpster primop)
    (dump dumpster (primop-name primop))))

(define generated-dump-encoder
  (lambda (dumpster name)
    (dump dumpster (name->qualified name (fluid $cenv-for-dump)))))

(define closure-dump-encoder
  (lambda (dumpster closure)
    (dump-template dumpster (closure-template closure))
    (dump dumpster (closure-env closure))))

(define (dump-template dumpster template)
  (dump dumpster (template-ref template 0)) ; byte code vector
  (dump dumpster (template-ref template 1)) ; other code vector
  (dump-debug-data dumpster (template-ref template 2)) ; debug data
  ;; Don't dump the package key.
  (let ((len (template-length template)))
    (dump dumpster len)
    (do ((i 4 (+ i 1)))
        ((= i len))
      (dump dumpster (template-ref template i)
            dump-encoder-for-templates))))

(define (dump-debug-data dumpster ddata)
  (cond ((get-debug-data ddata)         ; Debug data may be tabulated.
         => (lambda (ddata)
              (dump dumpster #t)
              (dump dumpster (debug-data-name ddata))
              (dump-debug-data dumpster (debug-data-parent ddata))
              (dump dumpster (debug-data-env-maps ddata))
              (dump dumpster (debug-data-source ddata))))
        (else
         (dump dumpster #f))))

(define (dump-encoder-for-templates obj)
  (cond ((location? obj) (values #f 'location location-dump-encoder))
        ((template? obj) (values #t 'template template-dump-encoder))
        (else            (usual-dump-encoder obj))))

(define location-dump-encoder
  (lambda (dumpster loc)
    (cond ((table-ref (fluid $location-table) loc)
           => (lambda (qname.type)
                (dump dumpster (location-id loc))
                (dump dumpster (car qname.type))
                (dump dumpster (cdr qname.type))))
          (else
           (error "dumping unrecognized location" loc)))))

(define template-dump-encoder
  (lambda (dumpster template)
    (dump-template dumpster template)))



(define (retrieve-and-load-package package call-with-retriever)
  (if (not (package-loaded? package))
      (let ((template
             (retrieve-package-template package call-with-retriever)))
        (with-interaction-environment package
          (lambda ()
            (invoke-closure (make-closure template
                                          (package-uid package)))))
        (set-package-loaded?! package #t)
        (walk-population check-structure
                         (package-clients package)))))

(define (retrieve-package-template package call-with-retriever)
  (call-with-retriever dump-decoder-for-packages
    (lambda (retriever)
      (if (not (= (retrieve-integer retriever)
                  package-dump-level))
          (error "package dump format version mismatch"
                 package
                 '(try recompiling)))
      (let-fluids $package package
                  $generated-names (make-cell '())
        (lambda ()
          (retrieve-bindings! retriever package)
          (retrieve-template retriever))))))

(define $package         (make-fluid #f))
(define $generated-names (make-fluid #F))

(define (retrieve-bindings! retriever package)
  (let loop ()
    (cond ((retrieve retriever)
           => (lambda (name)
                (let* ((type (sexp->type (retrieve retriever) #t))
                       (static (maybe-retrieve-transform retriever
                                                         package)))
                  (package-define! package name type #f static))
                (loop))))))

(define (maybe-retrieve-transform retriever package)
  (cond ((retrieve retriever)
         => (lambda (id)
              (let* ((type (sexp->type (retrieve retriever) #t))
                     (proc (retrieve retriever))
                     (aux-names (retrieve retriever))
                     (source (retrieve retriever)))
                (receive (proc.aux-names source)
                         (if (eq? proc 'inline-transform)
                             (values (inline-transform source
                                                       aux-names)
                                     `(inline-transform ',source
                                                        ',aux-names))
                             (values (cons proc aux-names) source))
                  (make-transform proc.aux-names
                                  package
                                  type
                                  source
                                  id)))))
        (else #f)))

(define (dump-decoder-for-packages key)
  (cond ((eq? key 'primop)       primop-dump-decoder)
        ((eq? key 'generated) generated-dump-decoder)
        ((eq? key 'closure)     closure-dump-decoder)
        (else (usual-dump-decoder key))))

(define primop-dump-decoder
  (lambda (retriever port register!)
    (get-primop (retrieve retriever))))

(define generated-dump-decoder
  (lambda (retriever port register!)
    (qualified->name (retrieve retriever))))

(define closure-dump-decoder
  (lambda (retriever port register!)
    (let ((closure (make-closure #f #f)))
      (register! closure)
      (set-closure-template! closure (retrieve-template retriever))
      (set-closure-env!      closure (retrieve retriever))
      closure)))

;;; This is used for defining setters for fields that don't have them
;;; built-in, namely closures & weak pointers.  This uses explicit
;;; renaming macros, not SYNTAX-RULES, because the LAP transducer can't
;;; handle the generated names in the output of SYNTAX-RULES macros.
;;; This is used locally to this module anyway, so it doesn't need to
;;; be hygienic.
;;;
;;; This is kind of gross, and it might be better placed in the object
;;; dumper itself.

(define-syntax define-stob-setter
  (lambda (form rename compare)
    (let ((name      (cadr   form))
          (stob-type (caddr  form))
          (index     (cadddr form)))
      `(DEFINE ,name
         (LAP ,name ()
           (PROTOCOL 2)                 ; Two arguments.
           (POP)
           (STORED-OBJECT-SET! ,stob-type ,index
                               0)       ; The 0 tells it not to log in
           (RETURN))))))                ;   the current proposal.

(define-stob-setter set-closure-template! closure 0)
(define-stob-setter set-closure-env!      closure 1)

(define (retrieve-template retriever)
  (let* ((code (retrieve retriever))
         (byte-code (retrieve retriever))
         (ddata (retrieve-debug-data retriever))
         (len (retrieve retriever))
         (template (make-template len #f)))
    (template-set! template 0 code)
    (template-set! template 1 byte-code)
    (template-set! template 2 ddata)
    (template-set! template 3 (package-uid (fluid $package)))
    (do ((i 4 (+ i 1)))
        ((= i len) template)
      (template-set! template i
                     (retrieve retriever
                               dump-decoder-for-templates)))))

(define (retrieve-debug-data retriever)
  (if (retrieve retriever)
      (let* ((name (retrieve retriever))
             (parent (retrieve-debug-data retriever))
             (env-maps (retrieve retriever))
             (source (retrieve retriever)))
        ;; Use a temporary, new debug data record to get a new uid;
        ;; then make a debug data record with the fields we retrieved
        ;; and this new uid.  Alternatively, this could create a debug
        ;; data record with NEW-DEBUG-DATA and then initialize it with
        ;; SET-DEBUG-DATA-ENV-MAPS! & SET-DEBUG-DATA-SOURCE!.
        (make-debug-data (debug-data-uid (new-debug-data #f #f))
                         name
                         parent
                         env-maps
                         source))
      #f))

(define (dump-decoder-for-templates key)
  (cond ((eq? key 'location) location-dump-decoder)
        ((eq? key 'template) template-dump-decoder)
        (else (usual-dump-decoder key))))

(define location-dump-decoder
  (lambda (retriever port register!)
    (let* ((id (retrieve retriever))
           (qname (retrieve retriever))
           (type (retrieve retriever))
           (loc (find-location qname
                               (sexp->type type #t)
                               (fluid $package))))
      (register! loc)
      loc)))

(define (find-location qname want-type package)
  (let ((name (qualified->name qname)))
    (cond ((package-lookup package name)
           => (lambda (binding)
                (check-binding-type binding want-type package qname)
                (binding-place binding)))
          (else
           (error "unable to retrieve location for name"
                  qname name want-type package)))))

(define (check-binding-type binding want-type package qname)
  (let ((real-type (binding-type binding)))
    (if (not (or (equal? want-type usual-variable-type)
                 (if (pair? want-type)
                     (and (pair? real-type)
                          (eq? (car want-type) 'variable)
                          (eq? (car real-type) 'variable)
                          (same-type? (cadr want-type)
                                      (cadr real-type)))
                     (and (not (pair? real-type))
                          (same-type? want-type real-type)))))
        (warn "type inconsistency in dumped binding"
              package qname
              `(compile-time type was ,want-type)
              `(load-time type is ,real-type)))))

;;; Taken roughly from opt/inline.scm, modified for our purposes.

(define (qualified->name name)
  (if (qualified? name)
      (let recur ((name name))
        (let* ((parent (qualified-parent-name name))
               (parent (if (qualified? parent)
                           (recur parent)
                           (find-generated-name parent))))
          (generate-name (qualified-symbol name)
                         (get-qualified-env (generated-env  parent)
                                            (generated-name parent))
                         parent)))
      name))

(define (get-qualified-env env name)
  (let ((probe (generic-lookup env name)))
    (if (and (binding? probe)
             (transform? (binding-static probe)))
        (transform-env (binding-static probe))
        (error "invalid qualified reference"
               env name probe))))

(define (find-generated-name name)
  (let* ((cell (fluid $generated-names))
         (alist (cell-ref cell)))
    (cond ((assq name alist) => cdr)
          (else
           (let ((generated
                  (generate-name name
                                 (package->environment
                                  (fluid $package))
                                 #f)))
             (cell-set! cell (cons (cons name generated) alist))
             generated)))))

(define template-dump-decoder
  (lambda (retriever port register!)
    (retrieve-template retriever)))
