;;; Load this into the exec package after fixing the following
;;; translations.  Then, to test FASL dumping, create a file foo.fasl
;;; in the current directory and evaluate (FASL-DUMP-TEST) in the exec
;;; package; to test FASL loading, just run (LOAD-FASL "foo.fasl") in
;;; the user package.  To test package dumping, run (PACKAGE-DUMP-TEST)
;;; in the exec package; to test package loading, simply evaluate
;;; (PACKAGE-LOAD-TEST) in the exec package.  For a really big test,
;;; get Essence, edit the call to TRANSLATE in DUMP-ESSENCE-FASL, and
;;; just run (DUMP-ESSENCE-FASL) in the exec package.

(translate "=riatables/" "../riatables/s48-riatables/")
(translate "=objdumper/" "../objdumper/s48-objdumper/")
(translate "=fasl/" "./")

(config '(load "=riatables/scheme/interfaces.scm"
               "=riatables/scheme/packages.scm"
               "=objdumper/scheme/interfaces.scm"
               "=objdumper/scheme/packages.scm"
               "=fasl/interfaces.scm"
               "=fasl/packages.scm"))

(load-package 'fasl-commands)

(config '(load "=fasl/test-packages.scm"))

(define (dump-essence-fasl)
  (translate "=essence/" "../../../essence-1.0/")
  (config '(load "=essence/genext/genext-packages.scm"
                 "=essence/interfaces.scm"
                 "=essence/packages.scm"
                 "=essence/cps/cps-lr-genext.config.scm"
                 "=essence/generator-packages.scm"))
  ;; Must wrap this in an EXEC because we need it compiled after we've
  ;; loaded the FASL-COMMANDS package & the Essence modules, but that
  ;; won't happen until after this file is *loaded* (& the previous
  ;; LOAD evaluated).
  (exec '(dump-fasl "essence.fasl"
                    (cps-lr cps-lr-generate grammar stream)
                    auxiliary cogen-boxops cogen-completers
                    cogen-directives cogen-gensym cogen-globals
                    cogen-library cogen-memo-standard cogen-record
                    cogen-residual cogen-specialize cps-lr-genext
                    define-data lr-spectime pgg-library pgg-specialize
                    scc-union shift-reset)))

(define (dump-sort-lib-fasl)
  (translate "=sort/" "../../scheme48-1.2/scheme/sort/")
  ;; These are already loaded in the usual Scheme48 image, but we need
  ;; VECTOR-UTILS & VECTOR-INSERTION-SORT-INTERNAL, which are not
  ;; exposed for whatever reason.  It's simpler to just reload these
  ;; all than to extract the two structures from the exposed ones.
  (config '(load "=sort/interfaces.scm"
                 "=sort/packages.scm"))
  (config '(run (define-structure sort (export sort-list sort-list!)
                  (open scheme-level-2 list-merge-sort)
                  (optimize auto-integrate)
                  (begin (define (sort-list list <)
                           (list-merge-sort < list))
                         (define (sort-list! list <)
                           (list-merge-sort! < list))))))
  (exec '(dump-fasl "sort.fasl"
                    (sorting sorted delete-neighbor-duplicates
                     binary-searches list-merge-sort vector-merge-sort
                     vector-heap-sort vector-insertion-sort
                     sort)
                    vector-utils vector-insertion-sort-internal)))

(define (fasl-dump-test)                ; No longer much point...
  (exec '(dump-fasl "foo.fasl" (foo))))

(define (package-dump-test)
  (*test-package 'compile-and-dump-package-to-file))

(define (package-load-test)
  (*test-package 'retrieve-and-load-package-from-file))

(define (*test-package op)
  (new-package)
  (open 'package-commands-internal 'packages 'package-dumping)
  (run `(,op (structure-package (get-structure 'foo)) "foo.fasl"))
  (user))
