;;;;;; FASL library for Scheme48                         -*- Scheme -*-
;;;;;; Package definitions

;;; Taylor Campbell wrote this code; he places it in the public domain.

(define-structure package-dumping package-dumping-interface
  (open scheme
        receiving
        destructuring
        signals
        (subset util                    (fold))
        fluids
        tables
        cells
        object-dump
        (subset weak                    (walk-population))
        (subset interfaces              (make-simple-interface))
        (subset packages                (make-structure
                                         make-simple-package package?
                                         package->environment
                                         package-uid
                                         package-lookup package-define!
                                         generic-lookup
                                         link!))
        (subset packages-internal       (package-name
                                         package-opens
                                         package-loaded?
                                         set-package-loaded?!
                                         package-definitions
                                         package-clients))
        (subset environments            (with-interaction-environment))
        primops
        templates
        closures
        assembler                       ; to implement SET-CLOSURE-ENV!
        (subset vm-exposure             (invoke-closure))
        records                         ; for RECORD-REF for transforms
        locations
        bindings
        (subset names                   (generate-name generated?
                                         generated-env
                                         generated-name
                                         name->qualified qualified?
                                         qualified-parent-name
                                         qualified-symbol))
        (subset inline                  (inline-transform))
        (subset meta-types              (syntax-type vector-type
                                         usual-variable-type
                                         type->sexp sexp->type
                                         same-type?))
        (subset compiler                (compile-forms))
        optimizer
        (subset scan-package            (package-optimizer-names
                                         package-source
                                         check-structure))
        (subset syntactic               (expand scan-forms
                                         expand-scanned-form))
        (subset compiler-envs           (make-compiler-env
                                         bind-source-file-name
                                         environment-define!
                                         environment-macro-eval))
        transforms
        debug-data
        (subset debug-data-internal     (new-debug-data
                                         make-debug-data))
        usual-macros
        (subset nodes                   (make-node
                                         get-operator
                                         operator-type
                                         operators-table))
        (subset primops                 (get-primop walk-primops)))
  (files dump-pkg))

(define-structure fasl fasl-interface
  (open scheme
        receiving
        object-dump
        filenames
        fluids
        riatables
        signals
        srfi-2                          ;and-let*
        destructuring
        interfaces
        meta-types
        weak
        (subset extended-ports          (call-with-string-output-port
                                         make-string-input-port))
        package-dumping
        (subset packages                (structure-interface
                                         structure-package
                                         note-structure-name!
                                         make-structure structure?
                                         make-modified-structure
                                         make-package package?
                                         package-uid
                                         package-lookup))
        (subset packages-internal       (structure-name
                                         package-name
                                         package-opens
                                         package-accesses
                                         package-unstable?
                                         package-file-name
                                         package-clauses
                                         package-clients
                                         package-loaded?
                                         set-package-loaded?!))
        compile-packages
        (subset bindings                (binding-type binding-place))
        locations
        (subset environments            (with-interaction-environment
                                         environment-define!
                                         environment-ref))
        (subset vm-exposure             (invoke-closure))
        (subset closures                (make-closure))
        (subset package-commands-internal
                                        (config-package
                                         get-structure))
        (subset scan-package            (collect-packages
                                         check-structure)))
  (files fasl))

(define-structure fasl-commands (export)
  (open scheme
        receiving
        handle
        signals
        (subset filenames               (translate
                                         file-name-directory))
        fasl
        package-dumping
        (subset environments            (environment-define!
                                         environment-ref))
        (subset command-processor       (define-user-command-syntax
                                         user-command-environment
                                         config-package))
        (subset packages                (structure-package))
        (subset packages-internal       (package-unstable?
                                         package-file-name))
        (subset package-commands-internal
                                        (config-package
                                         get-structure)))
  (files commands))
